%% @doc A functional FIFO queue.
%% @author Karl Marklund <karl.marklund@it.uu.se>
%% @author Henrik Alfken <henrik.schulze.0353@student.uu.se>
-module(fifo).
-export_type([fifo/0]).
%% The data type of a FIFO queue:
-opaque fifo()::{fifo, list(), list()}.
-export([new/0, size/1, push/2, pop/1, empty/1]).

%% To use EUnit we must include this:
-include_lib("eunit/include/eunit.hrl").

%% @doc Creates an empty FIFO queue.
%% Represent the FIFO using a 3-tuple {fifo, In, Out};
%% where In and Out are lists.
%% === Example ===
%% <div class="example">```
%% > F0 = fifo:new().
%% {fifo,[],[]}'''
%% </div>
-spec new() -> fifo().
new() -> {fifo,[],[]}.

%% @doc Gets the size of the queue.
%% === Example ===
%% <div class="example">```
%% > fifo:size({fifo,[],[]}).
%% 0
%% > fifo:size({fifo,[b,a],[]}).
%% 2
%% > fifo:size({fifo,[],[a,b,c]}).
%% 3'''
%% </div>
-spec size(Fifo) -> integer() when Fifo::fifo().
size({fifo, In, Out}) -> length(In) + length(Out).

%% @doc Pushes an element to the queue.
%% Adds the new value to the head of In.
%% === Example ===
%% <div class="example">```
%% > F1a = fifo:push({fifo,[],[]},a).
%% {fifo,[c,b,a],[]}
%% > F2ab = fifo:push(F1a,b).
%% {fifo,[c,b,a],[]}
%% > F3 = fifo:push({fifo,[b,a],[]},c).
%% {fifo,[c,b,a],[]}
%% > fifo:push({fifo,[b,a],[]},"Hello!").
%% {fifo,[c,b,a],[]}'''
%% </div>
-spec push(Fifo::fifo(),X::any()) -> fifo().
push({fifo, In, Out}, X) -> {fifo,[X|In], Out}.

%% @doc Pops an element from the queue.
%% === Example ===
%% <div class="example">```
%% > {_,F2bc} = fifo:pop({fifo,[],[a,b,c]}).
%% {a,{fifo,[],[b,c]}}
%% > {_,F1b} = fifo:pop({fifo,[b,a],[]}).
%% {a,{fifo,[],[b]}}
%% > fifo:pop(F2ab).
%% {b,{fifo,[],[]}}
%% > fifo:pop(F1b).
%% {b,{fifo,[],[]}}'''
%% </div>
%% @throws 'empty fifo'
-spec pop(Fifo::fifo()) -> {any(),fifo()}.
pop({fifo,[],[]}) -> erlang:error('empty fifo');
%% To make pop fast we want to pop of the head of the Out list.
pop({fifo, In,[H|T]}) -> {H,{fifo, In,T}};
%% When Out is empty, we must take a performance penalty, using the
%% reverse of In as the new Out and an empty lists as the new In,
%% then pop as usual.
pop({fifo,In,[]}) -> pop({fifo,[],lists:reverse(In)}).

%% @doc Checks if the queue is empty.
-spec empty(Fifo) -> boolean() when Fifo::fifo().
empty({fifo,[],[]}) -> true;
empty({fifo, _, _}) -> false.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%% Eunit test cases  %%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% EUnit adds the fifo:test() function to this module.
%% All functions with names ending wiht _test() or _test_() will be
%% called automatically by fifo:test()
new_test_() ->
    [?_assertEqual({fifo,[],[]}, new()),
     ?_assertMatch(0, fifo:size(new())),
     ?_assertException(error, 'empty fifo', pop(new()))].

push_test() ->
    push(new(), a).

push_pop_test() ->
    ?assertMatch({a,_}, pop(push(new(), a))).

f1() ->
    push(push(push(new(), foo), bar), "Ahloa!").

size_test_() ->
    F1 = f1(),
    F2 = push(F1, any),
    {_, F3} = fifo:pop(F2),
    [?_assertMatch(3, fifo:size(F1)),
     ?_assertMatch(4, fifo:size(F2)),
     ?_assertMatch(3, fifo:size(F3))].

push_test_() ->
    F1 = f1(),
    F2 = push(f1(), last),
    [ ?_assertMatch(1, fifo:size(fifo:push(fifo:new(), a))),
      ?_assertEqual(fifo:size(F1) + 1, fifo:size(F2))].

empty_test_() ->
    F = f1(),
    {_, F2} = pop(F),
    {_, F3} = pop(F2),
    {_, F4} = pop(F3),
    [?_assertMatch(true, empty(new())),
     ?_assertMatch(false, empty(F)),
     ?_assertMatch(false, empty(F2)),
     ?_assertMatch(false, empty(F3)),
     ?_assertMatch(true, empty(F4))].
